@@ -67,33 +67,49 @@ def _get_cols(linear, monteCarlo, parameterize):
     '''Helper function to create the names of the data to output'''
     save_columns = ['Sample',
                     'Raw date',
-                    'Mean raw date',
-                    'Linear raw uncertainty',
-                    ' +68% CI raw',
-                    ' -68% CI raw',
+                    'Linear raw 1σ uncertainty',
+                    'MC average 68% CI, raw',
+                    'MC +68% CI, raw',
+                    'MC -68% CI, raw',
+                    'Linear raw 2σ uncertainty',
+                    'MC average 95% CI, raw',
+                    'MC +95% CI, raw',
+                    'MC -95% CI, raw',
                     'Corrected date',
-                    'Mean corrected date',
-                    'Linear corrected uncertainty',
-                    ' +68% CI corrected',
-                    ' -68% CI corrected',
+                    'Linear corrected 1σ uncertainty',
+                    'MC average 68% CI, corrected',
+                    'MC +68% CI, corrected',
+                    'MC -68% CI, corrected',
+                    'Linear corrected 2σ uncertainty',
+                    'MC average 95% CI, corrected',
+                    'MC +95% CI, corrected',
+                    'MC -95% CI, corrected',
                     'Number of Monte Carlo simulations']
     if monteCarlo and not linear:
         save_columns = ['Sample',
                         'Raw date',
-                        'Mean raw date',
-                        ' +68% CI raw',
-                        ' -68% CI raw',
+                        'MC average 68% CI, raw',
+                        'MC +68% CI, raw',
+                        'MC -68% CI, raw',
+                        'MC average 95% CI, raw',
+                        'MC +95% CI, raw',
+                        'MC -95% CI, raw',
                         'Corrected date',
-                        'Mean corrected date',
-                        ' +68% CI corrected',
-                        ' -68% CI corrected',
+                        'MC average 68% CI, corrected',
+                        'MC +68% CI, corrected',
+                        'MC -68% CI, corrected',
+                        'MC average 95% CI, corrected',
+                        'MC +95% CI, corrected',
+                        'MC -95% CI, corrected',
                         'Number of Monte Carlo simulations']
     elif linear and not monteCarlo:
         save_columns = ['Sample',
                         'Raw date',
-                        'Linear raw uncertainty',
+                        'Linear raw 1σ uncertainty',
+                        'Linear raw 2σ uncertainty',
                         'Corrected date',
-                        'Linear corrected uncertainty']
+                        'Linear corrected 1σ uncertainty',
+                        'Linear corrected 2σ uncertainty']
     elif not monteCarlo and not linear:
         save_columns = ['Sample',
                         'Raw date',
@@ -101,12 +117,9 @@ def _get_cols(linear, monteCarlo, parameterize):
     # If parameterization is chosen, add the relevant columns
     # at the end of the raw and corrected data output
     if parameterize:
-        ins_idx = save_columns.index(' -68% CI raw')
-        add_cols = ['raw fit a','raw fit u','raw fit s']
-        save_columns = save_columns[:ins_idx+1]+add_cols+save_columns[ins_idx+1:]
-        ins_idx = save_columns.index(' -68% CI corrected')
-        add_cols = ['corrected fit a','corrected fit u','corrected fit s']
-        save_columns = save_columns[:ins_idx+1]+add_cols+save_columns[ins_idx+1:]
+        add_cols = ['Hist raw fit a','Hist raw fit u','Hist raw fit s',
+                    'Hist corrected fit a','Hist corrected fit u','Hist corrected fit s']
+        save_columns = save_columns+add_cols
     return save_columns
 
 def _load_file(file, measured_U235, sheet):
@@ -156,12 +169,31 @@ def _load_file(file, measured_U235, sheet):
         except AssertionError:
             return None
     
+    # build list of expected correlation coefficient columns
+    corrs = [['r 238U-235U', 'r 235U-238U'],
+             ['r 238U-232Th', 'r 232Th-238U'],
+             ['r 238U-147Sm', 'r 147Sm-238U'],
+             ['r 235U-232Th', 'r 232Th-235U'],
+             ['r 235U-147Sm', 'r 147Sm-235U'],
+             ['r 232Th-147Sm', 'r 147Sm-232Th'],
+             ['r 238Ft-235Ft', 'r 235Ft-238Ft'],
+             ['r 238Ft-232Ft', 'r 232Ft-238Ft'],
+             ['r 238Ft-147Ft', 'r 147Ft-238Ft'],
+             ['r 235Ft-232Ft', 'r 232Ft-235Ft'],
+             ['r 235Ft-147Ft', 'r 147Ft-235Ft'],
+             ['r 232Ft-147Ft', 'r 147Ft-232Ft']]
+    
     # create a list of the column headers and the name of following column
     # to get the uncertainty columns as well
     col_load = ['Sample']
     for c in cols:
         col_load.append(c)
         col_load.append(data_load.columns[data_load.columns.get_loc(c)+1])
+    # Add correlation coefficient columns
+    for n in corrs:
+        for co in n:
+            if co in data_load.columns:
+                col_load.append(co)
     
     # Create list of column header names to load
     final_cols = ['Sample']
@@ -172,10 +204,37 @@ def _load_file(file, measured_U235, sheet):
     # Restrict data only to relevant data and shorten the
     # column names for easier handling down the road
     data = data_load[col_load]
-    data.columns = final_cols
+    data.rename(columns={k:v for k, v in zip(col_load, final_cols)}, inplace = True)
     short_names = {name:name.replace('mol ','') for name in data.columns if 'mol ' in name}
     data = data.rename(columns=short_names)
     
+    # Identify which correlation coefficients are present
+    cor_cols = []
+    missing_cors = []
+    for c in corrs:
+        found = False
+        for d in c:
+            if d in data:
+                cor_cols.append(d)
+                found = True
+        if not found:
+            missing_cors.append(c[0].split(' ')[-1])
+    
+    # Convert correlation coefficients to Covariance (if present)
+    order_pref = ['238U', '235U','232Th','147Sm', '238Ft', '235Ft', '232Ft', '147Ft']
+    for c in cor_cols:
+        # Put isotopes and Fts in consistent order
+        isos = c.split(' ')[-1].split('-')
+        if order_pref.index(isos[0]) > order_pref.index(isos[1]):
+            isos = isos[::-1]
+        # Convert correlation coefficients to Covariance
+        data[isos[0]+'-'+isos[1]] = data[u'\u00B1 '+isos[0]]*data[u'\u00B1 '+isos[1]]*data[c]
+        data.drop(c, axis=1, inplace=True)
+    # Add missing covariance values to dataframe, default to 0
+    data[missing_cors] = 0
+    
+    data.dropna(axis=0, how = 'all', inplace=True)
+    
     return data
 
 def _make_excel(save_out, save_columns, file, monteCarlo, precision_user, saveAs):
@@ -216,7 +275,10 @@ def _make_excel(save_out, save_columns, file, monteCarlo, precision_user, saveAs
     # Create a list to format for Excel
     colHead_row = "2:2"
     save_final = []
-    for s in save_out:
+    # Use the ordering of keys obtained in _get_cols to dictate the order of columns in the output
+    # since save_out is a dictionary this is both safer in terms of ordering and allowed columns
+    # to be more easily reordered in the future
+    for s in save_columns:
         save_final.append(save_out[s])
     save_final = list(zip(*save_final))
     save_final.insert(0, save_columns)
@@ -249,8 +311,12 @@ def _make_excel(save_out, save_columns, file, monteCarlo, precision_user, saveAs
             output.column_dimensions[utils.get_column_letter(cell.column)].width = 10
         if 'simulations' in cell.value:
             output.column_dimensions[utils.get_column_letter(cell.column)].width = 17
-        if 'Mean corrected' in cell.value:
-            output.column_dimensions[utils.get_column_letter(cell.column)].width = 14
+        if 'average CI' in cell.value:
+            output.column_dimensions[utils.get_column_letter(cell.column)].width = 12
+        if '% CI, corrected' in cell.value:
+            output.column_dimensions[utils.get_column_letter(cell.column)].width = 12
+        if 'Hist corrected' in cell.value:
+            output.column_dimensions[utils.get_column_letter(cell.column)].width = 13
     output.row_dimensions[3].height = 31
     
     try:
@@ -270,7 +336,10 @@ def _linear_propagation(nominal_t, data, measured_U235):
             data['4He'], nominal_t['raw date'], data[u'\u00B1 4He'],
             data['238U'], data['235U'], data['232Th'], data['147Sm'],
             U238_s=data[u'\u00B1 238U'], U235_s=data[u'\u00B1 235U'],
-            Th232_s=data[u'\u00B1 232Th'], Sm147_s=data[u'\u00B1 147Sm']
+            Th232_s=data[u'\u00B1 232Th'], Sm147_s=data[u'\u00B1 147Sm'],
+            U238_U235_v=data['238U-235U'], U238_Th232_v=data['238U-232Th'],
+            U238_Sm147_v=data['238U-147Sm'], U235_Th232_v=data['235U-232Th'],
+            U235_Sm147_v=data['235U-147Sm'], Th232_Sm147_v=data['232Th-147Sm']
             )
         linear_corr_uncertainty = date_uncertainty_with235(
             data['4He'], nominal_t['corrected date'], data[u'\u00B1 4He'],
@@ -279,14 +348,19 @@ def _linear_propagation(nominal_t, data, measured_U235):
             data[u'\u00B1 238U'], data[u'\u00B1 235U'],
             data[u'\u00B1 232Th'], data[u'\u00B1 147Sm'],
             data[u'\u00B1 238Ft'], data[u'\u00B1 235Ft'],
-            data[u'\u00B1 232Ft'], data[u'\u00B1 147Ft']
+            data[u'\u00B1 232Ft'], data[u'\u00B1 147Ft'],
+            data['238U-235U'], data['238U-232Th'], data['238U-147Sm'],
+            data['235U-232Th'], data['235U-147Sm'], data['232Th-147Sm'],
+            data['238Ft-235Ft'], data['238Ft-232Ft'], data['238Ft-147Ft'],
+            data['235Ft-232Ft'], data['235Ft-147Ft'], data['232Ft-147Ft']
             )
     else:
         linear_raw_uncertainty = date_uncertainty(
             data['4He'], nominal_t['raw date'], data[u'\u00B1 4He'],
             data['238U'], data['238U']/137.818, data['232Th'], data['147Sm'],
             U238_s=data[u'\u00B1 238U'], Th232_s=data[u'\u00B1 232Th'],
-            Sm147_s=data[u'\u00B1 147Sm']
+            Sm147_s=data[u'\u00B1 147Sm'], U238_Th232_v=data['238U-232Th'],
+            U238_Sm147_v=data['238U-147Sm'], Th232_Sm147_v=data['232Th-147Sm']
             )
         linear_corr_uncertainty = date_uncertainty(
             data['4He'], nominal_t['corrected date'], data[u'\u00B1 4He'],
@@ -294,13 +368,19 @@ def _linear_propagation(nominal_t, data, measured_U235):
             data['238Ft'], data['235Ft'], data['232Ft'], data['147Ft'],
             data[u'\u00B1 238U'], data[u'\u00B1 232Th'], data[u'\u00B1 147Sm'],
             data[u'\u00B1 238Ft'], data[u'\u00B1 235Ft'],
-            data[u'\u00B1 232Ft'], data[u'\u00B1 147Ft']
+            data[u'\u00B1 232Ft'], data[u'\u00B1 147Ft'],
+            data['238U-232Th'], data['238U-147Sm'], data['232Th-147Sm'],
+            data['238Ft-235Ft'], data['238Ft-232Ft'], data['238Ft-147Ft'],
+            data['235Ft-232Ft'], data['235Ft-147Ft'], data['232Ft-147Ft']
             )
     return {'raw unc': linear_raw_uncertainty, 'corr unc': linear_corr_uncertainty}
 
 def _sample_loop(save_out, sample_data, measured_U235, linear, monteCarlo,
-                 histograms, parameterize, decimals, precision):
+                 histograms, parameterize, decimals, precision): 
     '''Gets all relevant data for a single samle'''
+    # set up reject variable to keep track of whether to run MC
+    reject = False
+    
     # Create the U235 data
     U235 = sample_data['238U']/137.818
     if measured_U235:
@@ -311,30 +391,48 @@ def _sample_loop(save_out, sample_data, measured_U235, linear, monteCarlo,
                          sample_data['232Th'], sample_data['147Sm'],
                          sample_data['238Ft'], sample_data['235Ft'],
                          sample_data['232Ft'], sample_data['147Ft'])
+    # Reject MC run if nominal date produces NaN or impossibly
+    # old dates, as dates >1e10 yrs requires a long time to run
+    # and is extremely unlikely to be a legitimate date
+    for t in nominal_t.values():
+        if t > 1e10:
+            reject = True
+        elif np.isnan(t):
+            reject = True
     save_out['Raw date'].append(round(nominal_t['raw date']/1e6, decimals))
     save_out['Corrected date'].append(round(nominal_t['corrected date']/1e6, decimals))
     
     # Calls helper function _linear_propagation, which unpacks relevant
     # data and calls the correct functions for linear uncertainty propagation
     linear_uncertainty = _linear_propagation(nominal_t, sample_data, measured_U235)
+    
+    # Reject MC run if linear uncertainty produces NaN
+    for u in linear_uncertainty.values():
+        if np.isnan(u):
+            reject = True
+    
     if linear:
-        save_out['Linear raw uncertainty'].append(round(linear_uncertainty['raw unc']/1e6,decimals))
-        save_out['Linear corrected uncertainty'].append(round(linear_uncertainty['corr unc']/1e6,decimals))
+        save_out['Linear raw 1σ uncertainty'].append(round(linear_uncertainty['raw unc']/1e6,decimals))
+        save_out['Linear corrected 1σ uncertainty'].append(round(linear_uncertainty['corr unc']/1e6,decimals))
+        save_out['Linear raw 2σ uncertainty'].append(round((linear_uncertainty['raw unc']*2)/1e6,decimals))
+        save_out['Linear corrected 2σ uncertainty'].append(round((linear_uncertainty['corr unc']*2)/1e6,decimals))
     
-    #TODO set that If data don't result in valid date, skip Monte Carlo simulation
-    # if any(np.isnan(save_out[n] for n in save_out)):
-    #     monteCarlo = False
+    for k in sample_data:
+        if u'\u00B1' in k:
+            if sample_data[k]<0:
+                reject = True
     
-    if monteCarlo:
+    if monteCarlo and not reject:
         # Estimate the number of cycles needed to reach the requested precision
         s_est = linear_uncertainty['corr unc']
         mean_est = nominal_t['corrected date']
-        mc_number = s_est**2/(precision*mean_est)**2
+        mc_sd = precision*mean_est
+        # mc_number = s_est**2/(precision*mean_est)**2
+        mc_number = (2*mc_sd**2+s_est**2)/(2*mc_sd**2)
         if mc_number < 5:
             mc_number = 5
         else:
             mc_number = int(mc_number)
-        save_out['Number of Monte Carlo simulations'].append(mc_number)
         
         # Create U235 data for the monteCarlo function
         if not measured_U235:
@@ -345,8 +443,6 @@ def _sample_loop(save_out, sample_data, measured_U235, linear, monteCarlo,
             U235_s = sample_data[u'\u00B1 235U']
         
         # Call the monte carlo module
-        #TODO recognize if the monte carlo has too many samples removed for precision level
-        #TODO recognize nans in mc results and handle accordingly
         mc_results = monte_carlo(
             mc_number, sample_data['4He'], sample_data[u'\u00B1 4He'],
             sample_data['238U'], U235, sample_data['232Th'], sample_data['147Sm'],
@@ -355,29 +451,63 @@ def _sample_loop(save_out, sample_data, measured_U235, linear, monteCarlo,
             sample_data[u'\u00B1 232Th'], sample_data[u'\u00B1 147Sm'],
             sample_data[u'\u00B1 238Ft'], sample_data[u'\u00B1 235Ft'],
             sample_data[u'\u00B1 232Ft'], sample_data[u'\u00B1 147Ft'],
+            sample_data['238U-235U'], sample_data['238U-232Th'], sample_data['238U-147Sm'],
+            sample_data['235U-232Th'], sample_data['235U-147Sm'], sample_data['232Th-147Sm'],
+            sample_data['238Ft-235Ft'], sample_data['238Ft-232Ft'], sample_data['238Ft-147Ft'],
+            sample_data['235Ft-232Ft'], sample_data['235Ft-147Ft'], sample_data['232Ft-147Ft'],
             histogram=histograms, parameterize=parameterize
             )
                 
         # Put Monte Carlo results into data framework for saving
         for ft in ['raw', 'corrected']:
-            save_out['Mean '+ft+' date'].append(round(mc_results[ft+' date']['mean']/1e6,decimals))
-            save_out[' +68% CI '+ft].append(round(mc_results[ft+' date']['+68% CI']/1e6,decimals))
-            save_out[' -68% CI '+ft].append(round(mc_results[ft+' date']['-68% CI']/1e6,decimals))
+            if 1-mc_results[ft+' date']['cycles']/mc_number < precision:
+                if ft == 'corrected':
+                    save_out['Number of Monte Carlo simulations'].append(mc_number)
+                CI_68high = (mc_results[ft+' date']['+68% CI']-nominal_t[ft+' date'])
+                CI_68low = (nominal_t[ft+' date']-mc_results[ft+' date']['-68% CI'])
+                mean_68CI = np.average([CI_68high, CI_68low])
+                save_out['MC average 68% CI, '+ft].append(round(mean_68CI/1e6,decimals))
+                save_out['MC +68% CI, '+ft].append(round(CI_68high/1e6,decimals))
+                save_out['MC -68% CI, '+ft].append(round(CI_68low/1e6,decimals))
+                CI_95high = (mc_results[ft+' date']['+95% CI']-nominal_t[ft+' date'])
+                CI_95low = (nominal_t[ft+' date']-mc_results[ft+' date']['-95% CI'])
+                mean_95CI = np.average([CI_95high, CI_95low])
+                save_out['MC average 95% CI, '+ft].append(round(mean_95CI/1e6,decimals))
+                save_out['MC +95% CI, '+ft].append(round(CI_95high/1e6,decimals))
+                save_out['MC -95% CI, '+ft].append(round(CI_95low/1e6,decimals))
+                if histograms:
+                    mc_results[ft+' date']['histogram'][0] = np.around(mc_results[ft+' date']['histogram'][0]/1e6,decimals)
+                    mc_results[ft+' date']['histogram'][1] = mc_results[ft+' date']['histogram'][1]
+                    save_out[ft+' histogram'].append(mc_results[ft+' date']['histogram'])
+                    if parameterize:
+                        save_out['Hist '+ft+' fit a'].append(round(mc_results[ft+' date']['a'], decimals))
+                        try:
+                            save_out['Hist '+ft+' fit u'].append(round(mc_results[ft+' date']['u']/1e6,decimals))
+                            save_out['Hist '+ft+' fit s'].append(round(mc_results[ft+' date']['s']/1e6,decimals))
+                        except:
+                            save_out['Hist '+ft+' fit u'].append(mc_results[ft+' date']['u'])
+                            save_out['Hist '+ft+' fit s'].append(mc_results[ft+' date']['s'])
+            else:
+                reject = True
+    if reject:
+        for ft in ['raw', 'corrected']:
+            if ft == 'corrected':
+                save_out['Number of Monte Carlo simulations'].append('NaN')
+            save_out['MC average 68% CI, '+ft].append('NaN')
+            save_out['MC +68% CI, '+ft].append('NaN')
+            save_out['MC -68% CI, '+ft].append('NaN')
+            save_out['MC average 95% CI, '+ft].append('NaN')
+            save_out['MC +95% CI, '+ft].append('NaN')
+            save_out['MC -95% CI, '+ft].append('NaN')
             if histograms:
-                mc_results[ft+' date']['histogram'][0] = np.around(mc_results[ft+' date']['histogram'][0]/1e6,decimals)
-                mc_results[ft+' date']['histogram'][1] = mc_results[ft+' date']['histogram'][1]
-                save_out[ft+' histogram'].append(mc_results[ft+' date']['histogram'])
+                save_out[ft+' histogram'].append([np.array(['NaN']),np.array(['NaN'])])
                 if parameterize:
-                    save_out[ft+' fit a'].append(mc_results[ft+' date']['a'])
-                    try:
-                        save_out[ft+' fit u'].append(round(mc_results[ft+' date']['u']/1e6,decimals))
-                        save_out[ft+' fit s'].append(round(mc_results[ft+' date']['s']/1e6,decimals))
-                    except:
-                        save_out[ft+' fit u'].append(mc_results[ft+' date']['u'])
-                        save_out[ft+' fit s'].append(mc_results[ft+' date']['s'])
+                    save_out['Hist, '+ft+' fit a'].append('NaN')
+                    save_out['Hist, '+ft+' fit u'].append('NaN')
+                    save_out['Hist, '+ft+' fit s'].append('NaN')
     return save_out
 
-def hecalc_main(file=None, saveAs=None, percent_precision=0.01, decimals=5, measured_U235=False,
+def hecalc_main(file=None, saveAs=None, percent_precision=0.01, decimals=2, measured_U235=False,
                 monteCarlo=True, linear=True, histograms=False, parameterize=False):
     '''
     Read in a file with He, radionuclide, and Ft data and their uncertainties
@@ -511,7 +641,7 @@ def hecalc_main(file=None, saveAs=None, percent_precision=0.01, decimals=5, meas
         
         save_out = _sample_loop(save_out, sample_data, measured_U235, linear, monteCarlo,
                                 histograms, parameterize, decimals, precision)
-
+    
     # Create the excel workbook using the input parameters
     book = _make_excel(save_out, save_columns, file, monteCarlo, percent_precision, saveAs)
     
